'''
**Description:** Module to easily create standard repositories with CRUD operations.

Example to use:
```python
from mangust228.utils.repo import AsyncBaseRepo, AsyncBaseRepoFactory

# Note the inheritance in the class! It's important for correct hinting.
class UserRepo(AsyncBaseRepo[UserModel]):
    model = UserModel

class Repository(AsyncBaseRepoFactory):
    session = session_maker
    # CAUTION: Use only annotation
    user: UserRepo 
    another_repo: AnotherRepo

async with Repository() as repo:
    user = await repo.user.add(name="Ivan", surname="Ivanov")
    user_id = user.id 
    await repo.user.count(surname="Ivanov")
    await repo.user.update_by_id(user_id, name="Sergey")
    await repo.user.get_one_or_none(surname="Ivanov")
    ...
```
'''

from .base_async_repo import AsyncBaseRepo
from .factory_async_repo import AsyncBaseRepoFactory



'''
My UserAgentRotator... because another libriries get old user-agent


Example:
```python 

# get Random User-Agent
UaRandom.web()
# get User-Agent like Chrome
UaRandom.web("chrome") 
# get User-Agent like Firefox
UaRandom.web("firefox")
# get User-Agent like Safari
UaRadom.web("safari")

'''
from .ua_random import UaRandom



import csv 
import os 
from ._logger import get_logger

class _BaseSyncFileDesciptor:
    def __init__(self, path: str, fields_name: list[str], debug: bool=False):
        self.path = path
        self.fields_name = fields_name
        self._created = False 
        self._data: set[str] = set()
        self.logger = get_logger(self.__class__.__name__, debug)

    def _create_folder(self):
        if not self._created:
            folder_path = "/".join(self.path.split("/")[:-1])
            os.makedirs(folder_path, exist_ok=True)
            if not os.path.exists(self.path):
                with open(self.path, "w") as afp:
                    writer = csv.DictWriter(afp, self.fields_name)
                    writer.writeheader()
        self._created = True 
        self._get_history_from_file()

    def _get_history_from_file(self):
        with open(self.path, "r") as afp: 
            reader = csv.DictReader(afp, self.fields_name)
            afp.readline()
            for row in reader: 
                self._data.add(row["path"])
                
    def is_exist(self, path):
        return path in self._data
    
    def _write_data(self, data):
        if not self._created:
            self._create_folder()
        with open(self.path, "a+") as afp:
            writer = csv.DictWriter(afp, self.fields_name)
            writer.writerow(data)
        self._data.add(data["path"])


class _SuccessSyncFileDesciptor(_BaseSyncFileDesciptor):
    def add(self, path: str):
        data = {self.fields_name[0]:path}
        self._write_data(data)
        self.logger.debug(f'add success path/url: {path}')

class _WrongSyncFileDesciptor(_BaseSyncFileDesciptor):
    def add(self, path: str, exc: str):
        data = {self.fields_name[0]:path, self.fields_name[1]:exc}
        self._write_data(data)
        self.logger.warning(f'add wrong path/url: {path}')


class SyncParsedManager:
    wrong = _WrongSyncFileDesciptor("data/wrong.csv", ["path", "exc"])
    success = _SuccessSyncFileDesciptor("data/success.csv", ["path"])



import logging


def get_logger(name: str, debug: bool = False):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG if debug else logging.WARNING)
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(levelname)-8s[%(name)s] : %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False  # Отключение распространения логов вверх по иерархии
    return logger


'''
Class for checking if a file or URL has been processed. Creates two files:
- `data/wrong.csv` (if errors occurred)
- `data/success.csv` (if no errors occurred)

Example usage:

```python
from parsed_manager import AsyncParsedManager as ParsedManager

# Adds an erroneous record
await ParsedManager.wrong.add("url/path", reason) 
# Adds a successfully parsed record
await ParsedManager.success.add("url/path")  

if ParsedManager.wrong.is_exist("url/path"):
    pass  # This block is executed if the URL/path has been parsed before.
    
'''

from .async_manager import AsyncParsedManager
from .sync_manager import SyncParsedManager


from types import TracebackType
from typing import Any, Optional
from sqlalchemy.orm import Session, sessionmaker
from ._logger import get_logger
from .exceptions import NotInitializedRepo

class _LazyRepo:
    def __init__(self, repo_class: type):
        '''
        Descriptor class for lazy initialization of repositories.
        '''
        self.repo_class = repo_class

    def __set_name__(self, owner: type["SyncBaseRepoFactory"], name: str):
        self.attr_name = name

    def __get__(self, instance: "SyncBaseRepoFactory", owner: type["SyncBaseRepoFactory"]):
        if instance is None:
            raise NotInitializedRepo
        if self.attr_name not in instance.__dict__:
            instance.__dict__[self.attr_name] = self.repo_class(instance._session)
        return instance.__dict__[self.attr_name]

class _FactoryMeta(type):
    '''
    Metaclass to automatically initialize _LazyRepo descriptors based on class annotations.
    '''
    def __new__(cls, name: str, bases: tuple[type, ...], dct: dict):
        annotations: dict = dct.get("__annotations__", {})
        if annotations:
            repos = {k: v for k, v in annotations.items()}
            for repo_name, repo_class in repos.items():
                dct[repo_name] = _LazyRepo(repo_class)
        return super().__new__(cls, name, bases, dct)

class SyncBaseRepoFactory(metaclass=_FactoryMeta):
    '''
    Factory base class to manage repository instances and transactions.
    Example usage:
    ```python
    from connection import session
    class Repo(SyncBaseRepoFactory):
        session = session
        user: UserRepo
    with Repo() as repo:
        repo.user.add(data)
    with Repo(commit=False) as repo:
        user = repo.user.find({"id": 1})
    ```
    '''
    session: sessionmaker[Session]

    def __init__(self, commit: bool = True, debug: bool = False, **kwargs: Any):
        '''
        Initialize the factory with a session and optional commit control.
        :param commit: Whether to commit changes automatically (default: True).
        :param debug: Enable debug logging (default: False).
        :param kwargs: Additional arguments for session configuration.
        '''
        self._session = self.session(**kwargs)
        self.commit = commit
        self.logger = get_logger(self.__class__.__name__, debug)

    def __enter__(self):
        return self

    def __exit__(
            self,
            exc_type: Optional[type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]) -> Optional[bool]:
        try:
            if not exc_type:
                if self.commit:
                    self._session.commit()
                    self.logger.debug("changes commit")
            else:
                self._session.rollback()
                self.logger.exception("Exception occurred")
        finally:
            self._session.close()
        return None



class ExceptionUaRandom(Exception):
    msg = ""

    def __init__(self, msg: str | None = None):
        super().__init__(msg or self.msg)


class ExceptionInstance(ExceptionUaRandom):
    msg = "You couldn't make instance of class "

    def __init__(self, owner_name: str):
        super().__init__(f"{self.msg} {owner_name}")





class _BaseSaverException(Exception):
    msg = ""

    def __init__(self, msg: str | None = None):
        super().__init__(msg or self.msg)


class ExcExpectedNamesArguments(_BaseSaverException):
    msg = "Expected names for file, got nothing"



import json
import lzma
from typing import Any

import aiofiles
import aiofiles.os as aos

from ._base import BaseSaveManager


class AsyncSaveManager(BaseSaveManager):
    '''
    File saver

    Example usage:
    ```python
    saver = AsyncSaveManager(base_path="example", compress=True)
    path = await saver.save_html("this is content", "seller_id", 4, 5)
    print(path)  # "example/2024/05/29/22/seller_id_4_5.html.xz"
    ```

    Parameters
    ----------
    base_path : str, optional
        The base folder where all files will be saved, by default "data"
    add_uuid : bool, optional
        Option to add a UUID at the end of the file name to ensure unique file names, by default False
    compress : bool, optional
        Option to compress files using lzma when enabled, by default False
    debug : bool, optional
        Enable debug logging, by default False
    '''
    async def _save_async_file(self, content: str, path: str):
        if not self.compress:
            async with aiofiles.open(path, "w") as fp:
                await fp.write(content)
        else:  # lzma
            compressed_data = lzma.compress(content.encode())
            async with aiofiles.open(path, "wb") as compress_fp:
                await compress_fp.write(compressed_data)

    async def save_json(self, data: dict, *names: Any) -> str:
        file_name = self._get_json_file_name(names)
        content = json.dumps(data, ensure_ascii=False)
        folder_path = self.folder_path
        await aos.makedirs(folder_path, exist_ok=True)
        path = folder_path + file_name
        await self._save_async_file(content, path)
        self.logger.debug(f"success saved: {path}")
        return path

    async def save_html(self, content, *names: Any) -> str:
        file_name = self._get_html_file_name(names)
        folder_path = self.folder_path
        await aos.makedirs(folder_path, exist_ok=True)
        path = folder_path + file_name
        await self._save_async_file(content, path)
        self.logger.debug(f"success saved: {path}")
        return path







import random

from ._browsers import Browsers
from ._engines import Engines
from ._oses import Oses


class UaRandom:
    _web_browsers = ["chrome", "firefox", "safari"]

    @classmethod
    def web(cls, browser: str | None = None) -> str:
        if browser is None:
            browser = random.choice(cls._web_browsers)

        os_system = getattr(Oses, browser)
        version, engine = getattr(Engines, browser)
        browser_string = getattr(Browsers, browser)(version)

        rv = f"; rv:{version}" if random.choice([True, False]) else ""
        return f"Mozilla/5.0 ({os_system}{rv}) {engine} {browser_string}"



import random
from typing import Type

from ._descriptor import BaseDescriptor


class _Chrome(BaseDescriptor):
    name = "chrome"

    def __get__(self, instance: None, owner: Type):
        super().__get__(instance, owner)
        version = random.randint(99, 125)
        return version, f"AppleWebKit/537.{random.randint(11, 75)} (KHTML, like Gecko)"


class _Safari(BaseDescriptor):
    name = "safari"

    def __get__(self, instance: None, owner: Type):
        super().__get__(instance, owner)
        version = random.randint(10, 16)
        return version, f"AppleWebKit/{random.randint(601, 605)}/{random.randint(1, 8)}/{random.randint(1, 50)} (KHTML, like Gecko)"


class _Firefox(BaseDescriptor):
    name = "firefox"

    def __get__(self, instance: None, owner: Type):
        super().__get__(instance, owner)
        version = random.randint(99, 120)
        if random.choice([True, False]):
            return version, "Gecko/20100101"
        return version, f"Gecko/{version}"


class Engines:
    chrome = _Chrome()
    safari = _Safari()
    firefox = _Firefox()



import random

from ._descriptor import BaseDescriptor

OS_VERSIONS = {
    "chrome": [
        ("Windows NT 10.0; Win64; x64", .5),
        ("Windows NT 10.0", .3),
        ("Windows NT 6.1", .2),
        ("Macintosh; Intel Mac OS X 10_15_7", .2),
        ("X11; Ubuntu; Linux x86_64", .1),
        ("Windows NT 6.1; WOW64; Trident/7.0; AS", .3)
    ],
    "firefox": [
        ("Windows NT 10.0; Win64; x64", .3),
        ("Windows NT 10.0", .2),
        ("Windows NT 6.1", .3),
        ("X11; Ubuntu; Linux x86_64", .4),
        ("Windows NT 6.1; WOW64; Trident/7.0; AS", .3),
    ],
    "safari": [
        ("Macintosh; Intel Mac OS X 10_15_7", .5),
        ("Macintosh; Intel Mac OS X 10.15", .5)
    ]

}


class _BaseOs(BaseDescriptor):
    lst = []
    weight = []

    def __get__(self, instance: None, owner: type):
        super().__get__(instance, owner)
        return random.choices(self.lst, weights=self.weight)[0]


class _ChromeOs(_BaseOs):
    name = "chrome"
    lst = [i[0] for i in OS_VERSIONS["chrome"]]
    weight = [i[1] for i in OS_VERSIONS["chrome"]]


class _FirefoxOs(_BaseOs):
    name = "firefox"
    lst = [i[0] for i in OS_VERSIONS["firefox"]]
    weight = [i[1] for i in OS_VERSIONS["firefox"]]


class _SafariOs(_BaseOs):
    name = "safari"
    lst = [i[0] for i in OS_VERSIONS["safari"]]
    weight = [i[1] for i in OS_VERSIONS["safari"]]


class Oses:
    chrome = _ChromeOs()
    firefox = _FirefoxOs()
    safari = _SafariOs()



from httpx import Response as HttpxResponse

from ._config import Settings
from ._exceptions import ProxyManagerHttpException
from ._logger import get_logger
from ._schemas import ProxySchema


class BaseProxyManager:
    '''How to use: 
    '''
    def __init__(self, debug:bool=False, **config):
        '''
        Could be next variables in kwargs: 
            url: str
            api_key: str
            service_name: str 
            service_id: int 
            lock_time: int 
            logic: Literal["linear", "sum_history"]
            type_id: int 
            location_id: int 
            ignore_hours: int 
        '''
        self.settings = Settings(**config)
        self.logger = get_logger(self.__class__.__name__, debug)
        self._create_params()
        self._create_data()

    def _create_params(self):
        self._params = {
            "parsed_service": self.settings.service_name,
            "parsed_service_id": self.settings.service_id,
            "lock_time": self.settings.lock_time,
        }
        if (item := self.settings.location_id):
            self._params["location_id"] = item
        if (item := self.settings.type_id):
            self._params["type_id"] = item

    def _create_data(self):
        self._data = self.params
        if (item := self.settings.logic):
            self._data["logic"] = item
        if (item := self.settings.lock_time):
            self._data["lock_time"] = item
        if (item := self.settings.ignore_hours):
            self._data["ignore_hours"] = item
    
    @staticmethod
    def _raise_if_not_200(response: HttpxResponse) -> None:
        try:
            response.raise_for_status()
        except: 
            raise ProxyManagerHttpException(response.json())


    @property
    def params(self) -> dict:
        return self._params.copy()

    @property
    def data(self) -> dict:
        return self._data.copy()

    def _new_proxy(self, response: HttpxResponse) -> ProxySchema:
        data = response.json()
        proxy = ProxySchema(**data)
        self.logger.debug(f"get new {proxy}")
        return proxy



from typing import Literal

from dotenv import load_dotenv
from pydantic_settings import BaseSettings, SettingsConfigDict

load_dotenv()

class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="PROXY_SERVICE_")
    url: str
    api_key: str
    service_name: str
    service_id: int
    lock_time: int
    logic: Literal["linear", "sum_history"] = 'linear'
    logic_base_time: int | None = None
    location_id: int | None = None
    type_id: int | None = None
    ignore_hours: int | None = None



from ._exceptions import ExceptionInstance


class BaseDescriptor:
    name = "base"

    def __set__(self, instance: None, owner: type):
        if instance is not None:
            raise ExceptionInstance(owner.__class__.__name__)
        owner.__dict__[self.name] = self  # type: ignore

    def __get__(self, instance: None, owner: type):
        if instance is not None:
            raise ExceptionInstance(owner.__class__.__name__)



from httpx import AsyncClient

from ._base import BaseProxyManager
from ._schemas import ProxySchema


class AsyncProxyManager(BaseProxyManager):
    async def start(self): 
        self.client = AsyncClient(
            base_url=self.settings.url,
            headers={
                "X-API-KEY": self.settings.api_key
            })
        self.logger.info("session is opened")
    
    async def close(self):
        await self.client.aclose()
        self.logger.info("session is closed")
    
    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def get(self) -> ProxySchema:
        response = await self.client.get("/proxies/rotations/one", params=self.params)
        self._raise_if_not_200(response)
        return self._new_proxy(response)

    async def free(self, proxy: ProxySchema) -> None:
        response = await self.client.get(f"/proxies/rotations/free/{proxy.id}")
        self._raise_if_not_200(response)
        self.logger.info(f'{proxy} is free')

    async def change_without_error(self, proxy: ProxySchema) -> ProxySchema:
        data = self.data
        data["proxy_id"] = proxy.id
        response = await self.client.put(f"/proxies/rotations", json=data)
        self._raise_if_not_200(response)
        self.logger.info(f'{proxy} changed without error')
        return self._new_proxy(response)

    async def change_with_error(self, proxy: ProxySchema, reason: str) -> ProxySchema:
        data = self.data
        data["proxy_id"] = proxy.id
        data["reason"] = reason
        response = await self.client.patch("/proxies/rotations", json=data)
        self._raise_if_not_200(response)
        self.logger.info(f'{proxy} changed with error')
        return self._new_proxy(response)



from pydantic import BaseModel


class ProxySchema(BaseModel):
    id: int
    server: str
    port: int
    username: str
    password: str
    
    def __str__(self):
        return f"proxy-{self.id}"
    
    @property
    def req_conn(self):
        return {
            "http": f"http://{self.username}:{self.password}@{self.server}:{self.port}",
            "https": f"https://{self.username}:{self.password}@{self.server}:{self.port}"
            }
    
    @property
    def pw_conn(self):
        return {
            "server": f"https://{self.server}:{self.port}",
            "username": self.username,
            "password": self.password
        }
        


class _FactoryException(Exception):
    msg = ""
    def __init__(self):
        super().__init__(self.msg)


class NotInitializedRepo(_FactoryException):
    msg = "\n\nYou need to initialize the repo, like:\n\nasync with Repo() as repo:\n    repo.some_repo.add(data)\n"


class NotInitializedSession(_FactoryException):
    msg = "\n\nForgot to add async_session in Repo:\n\n class Repo(FactoryRepoBase):\n    __session = async_session\n\nProbably, __session is not async_sessionmaker"


from typing import Any, Sequence

from sqlalchemy import delete, func, insert, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase


class AsyncBaseRepo[M: DeclarativeBase]:
    '''
    Base repository for asynchronous operations with SQLAlchemy models.
    
    How to use: 
    ```python 
    from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
    
    engine = create_async_engine(url)
    async_session = async_sessionmaker(engine)
    
    class UserRepo(AsyncBaseRepo[MyAlchemyModel]):
        session = async_session
        
    class Repository(AsyncBaseRepoFactory):
        user: UserRepo
    ```
    '''
    model: type[M]

    def __init__(self, session: AsyncSession):
        self.session = session

    def add_by_model(self, model: M) -> None:
        ''' 
        Add model and return None.
        
        Example: 
        ```python
        async with Repository() as repo: 
            my_model = MyModel(**some_data)
            repo.user.add_by_model(my_model)
        ```
        :param model: An instance of the SQLAlchemy model to be added.
        :return: None
        '''
        self.session.add(model)
    
    async def add_by_kwargs(self, **kwargs: Any) -> M:
        ''' Add an item and return the SQLAlchemy model instance.
        
        :param kwargs: Fields and values for the new record.
        :return: The newly added SQLAlchemy model instance.
        
        Example: 
        
        ```python
        # add new user: 
        async with Repository() as repo: 
            repo.user.add(name="Ivan", surname="Ivanov")
        ```
        '''
        stmt = insert(self.model)\
            .values(**kwargs)\
            .returning(self.model)  # type: ignore
        result = await self.session.execute(stmt)
        return result.scalar_one()

    async def delete(self, **by_filter: Any) -> None:
        '''Delete records based on filter criteria.
        
        :param by_filter: Filter criteria as keyword arguments.
        :return: Number of rows deleted.
        
        Example:
        ```python
        async with Repository() as repo:
            # Delete user with id=5
            await repo.user.delete(id=5)
            # Delete users where name="Ivan"
            await repo.user.delete(name="Ivan")
        ```
        '''
        stmt = delete(self.model).filter_by(**by_filter)
        await self.session.execute(stmt)

    async def update_by_id(self, id: int, **new_values: Any) -> M | None:
        '''
        Update values by id.
        
        :param id: The ID of the record to update.
        :param new_values: Fields and new values for the record.
        :return: The updated SQLAlchemy model instance or None if not found.
        
        Example:
        ```python
        async with Repository() as repo:
            await repo.user.update_by_id(id=5, name="Ivan")
        ```
        '''
        stmt = update(self.model)\
            .values(**new_values)\
            .where(self.model.id == id).returning(self.model) # type: ignore  
        item = await self.session.execute(stmt)
        return item.scalar_one_or_none()

    async def get_one_or_none(self, **by_filter: Any) -> M | None:
        '''
        Get a single record based on filter criteria.
        
        :param by_filter: Filter criteria as keyword arguments.
        :return: The SQLAlchemy model instance or None if not found.
        
        Example:
        ```python
        async with Repository() as repo:
            user = await repo.user.get_one_or_none(id=3)
        # >> None
        ```
        Note: Be careful when using filters that couldn't return multiple rows.
        '''
        stmt = select(self.model).filter_by(**by_filter)
        result = await self.session.execute(stmt)
        return result.scalar_one_or_none()

    async def get_many(self,
                       limit: int | None = None,
                       offset: int | None = None,
                       **by_filter: Any) -> Sequence[M]:
        '''
        Get multiple records based on filter criteria.
        
        :param limit: Maximum number of records to return.
        :param offset: Number of records to skip.
        :param by_filter: Filter criteria as keyword arguments.
        :return: A sequence of SQLAlchemy model instances.
        
        Example:
        ```python
        async with Repository() as repo:
            users = await repo.user.get_many(role="moderator")
        ```
        '''
        stmt = select(self.model)\
            .filter_by(**by_filter)\
            .limit(limit)\
            .offset(offset)
        result = await self.session.execute(stmt)
        return result.scalars().all()
    
    async def count(self, **filter_by: Any) -> int:
        '''
        Get the count of rows in the table based on filter criteria.
        
        :param filter_by: Filter criteria as keyword arguments.
        :return: The number of rows that match the filter criteria.
        
        Example:
        ```python
        async with Repository() as repo:
            # Count documents where role="admin"
            count = await repo.user.count(role="admin")
            # Count all documents in the table
            total_count = await repo.user.count()
        ```
        '''
        stmt = select(func.count()).select_from(self.model).filter_by(**filter_by)
        result = await self.session.execute(stmt)
        return result.scalar_one()


from datetime import datetime
from nanoid import non_secure_generate

from ._exceptions import ExcExpectedNamesArguments
from ._logger import get_logger


class BaseSaveManager:
    def __init__(self, base_path: str = "data", add_uuid: bool = False, compress: bool = False, debug: bool= False):
        base_path = base_path.rstrip("/")
        self.base_path = base_path
        self.add_uuid = add_uuid
        self.compress = compress
        self.logger = get_logger(self.__class__.__name__, debug)

    @property
    def date_path(self):
        return datetime.now().strftime("/%Y/%m/%d/%H/")

    @property
    def folder_path(self):
        return self.base_path + self.date_path

    @property
    def uuid(self):
        if self.add_uuid:
            return "_" + str(non_secure_generate(size=10))
        return ''

    def _get_json_file_name(self, names: tuple):
        path = self._get_name_from_name(names) + ".json"
        return f"{path}.xz" if self.compress else path

    def _get_html_file_name(self, names: tuple):
        path = self._get_name_from_name(names) + ".html"
        return f"{path}.xz" if self.compress else path

    def _get_name_from_name(self, names: tuple):
        if len(names) == 0:
            raise ExcExpectedNamesArguments
        return "_".join([str(n) for n in names]) + self.uuid





class _ProxyManagerException(Exception):
    msg = ""
    def __init__(self, msg:str|None = None):
        super().__init__(msg or self.msg)
                 
class ProxyManagerHttpException(_ProxyManagerException):
    msg = "Proxies exist, but none are available"  
    


'''
### Example:

```python
from proxy_manager import SyncProxyManager, AsyncProxyManager

# Sync version:
with SyncProxyManager() as pm:
    proxy_1: ProxySchema = pm.get()
    proxy_2: ProxySchema = pm.change_with_error(new_proxy, reason="just a test")
    proxy_3: ProxySchema = pm.change_without_error(new_proxy)
    pm.free(proxy_3)  # Free proxy on the server

# Async version:
async with AsyncProxyManager() as pm:
    proxy_1: ProxySchema = await pm.get()
    proxy_2: ProxySchema = await pm.change_with_error(new_proxy, reason="just a test")
    proxy_3: ProxySchema = await pm.change_without_error(new_proxy)
    await pm.free(proxy_3)  # Free proxy on the server
```

### How to use response with different libraries:

```
# 1. For use with requests `req_conn`
response = requests.get("http://example.com", proxies=proxy.req_conn)

# 2. For use with httpx `httpx_conn`
response = httpx.get("http://example.com", proxies=proxy.httpx_conn)

# 3. For use with playwright `pw_conn`
context = await browser.new_context(proxy=proxy.pw_conn)

```

### .env EXAMPLE

```bash 
# required
PROXY_SERVICE_URL=http://255.255.255.255:8000
PROXY_SERVICE_API_KEY=XXXXXXXXXXXXXX
PROXY_SERVICE_SERVICE_NAME=test
PROXY_SERVICE_SERVICE_ID=1 
PROXY_SERVICE_LOCK_TIME=5

# not required
PROXY_SERVICE_LOGIC=linear
PROXY_SERVICE_LOGIC_BASE_TIME=10 
PROXY_SERVICE_IGNORE_HOURS=24
PROXY_SERVICE_TYPE_ID=1 
PROXY_SERVICE_LOCATION_ID=1
```

'''
from ._schemas import ProxySchema
from .async_manager import AsyncProxyManager
from .sync_manager import SyncProxyManager



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

extras = {
    'repo': ['sqlalchemy'],
    'parsed': ['aiofiles', 'aiocsv'],
    'proxy': ['httpx'],
    'saver': ['aiofiles', 'nanoid'],
    'ua': []
}

extras['all'] = sum(extras.values(), [])

setup(
    name="mangust228",
    version="0.8.23",
    description="Lazy utils each I use sometimes",
    author="mangust228",
    author_email="bacek.mangust@gmail.com",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mangustik228/proxy_manager',
    install_requires=[],
    extras_require=extras,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    test_suite='tests',
    python_requires='>=3.9'
)



import asyncio

import aiocsv
import aiofiles
import aiofiles.os as aos
from ._logger import get_logger

class _BaseAsyncFileDesciptor:
    def __init__(self, path: str, fields_name: list[str], debug:bool=False):
        self.path = path
        self.logger = get_logger(self.__class__.__name__, debug)
        self.fields_name = fields_name
        self._created = False 
        self._lock = asyncio.Lock()
        self._data: set[str] = set()

    async def _create_folder(self):
        async with self._lock:
            if not self._created:
                folder_path = "/".join(self.path.split("/")[:-1])
                await aos.makedirs(folder_path, exist_ok=True)
                if not await aos.path.exists(self.path):
                    async with aiofiles.open(self.path, "w") as afp:
                        writer = aiocsv.AsyncDictWriter(afp, self.fields_name)
                        await writer.writeheader()
            self._created = True 
            await self._get_history_from_file()

    async def _get_history_from_file(self):
        async with aiofiles.open(self.path, "r") as afp: 
            reader = aiocsv.AsyncDictReader(afp, self.fields_name)
            await afp.readline()
            async for row in reader: 
                self._data.add(row["path"])
                
    def is_exist(self, path):
        return path in self._data
    
    async def _write_data(self, data):
        if not self._created:
            await self._create_folder()
        async with aiofiles.open(self.path, "a+") as afp:
            writer = aiocsv.AsyncDictWriter(afp, self.fields_name)
            await writer.writerow(data)
        path = data["path"]
        self._data.add(path)
        return path 


class _SuccessAsyncFileDesciptor(_BaseAsyncFileDesciptor):
    async def add(self, path: str):
        data = {self.fields_name[0]:path}
        path = await self._write_data(data)
        self.logger.debug(f'add success path/url: {path}')

class _WrongAsyncFileDesciptor(_BaseAsyncFileDesciptor):
    async def add(self, path: str, exc: str):
        data = {self.fields_name[0]:path, self.fields_name[1]:exc}
        path = await self._write_data(data)
        self.logger.warning(f'add wrong path/url: {path}')

class AsyncParsedManager:
    wrong = _WrongAsyncFileDesciptor("data/wrong.csv", ["path", "exc"])
    success = _SuccessAsyncFileDesciptor("data/success.csv", ["path"])


import random


class Browsers:
    @classmethod
    def safari(cls, version):
        return f"Safari/{version}.{random.randint(1, 15)}"

    @classmethod
    def chrome(cls, version):
        return f"Chrome/{version}.0.9999.{random.randint(0, 999)}"

    @classmethod
    def firefox(cls, version):
        return f"Firefox/{version}.0"



import json
import lzma
import os
from typing import Any

from ._base import BaseSaveManager


class SyncSaveManager(BaseSaveManager):
    '''
    File saver

    Example usage:
    ```python
    saver = SyncSaveManager(add_uuid=True)
    path = saver.save_json({"hello": "world"}, 5, 3, "daily")
    print(path)  # "data/2024/05/29/22/5_3_daily.json"
    ```

    Parameters
    ----------
    base_path : str, optional
        The base folder where all files will be saved, by default "data"
    add_uuid : bool, optional
        Option to add a UUID at the end of the file name to ensure unique file names, by default False
    compress : bool, optional
        Option to compress files using lzma when enabled, by default False
    debug : bool, optional
        Enable debug logging, by default False
    '''
    def _save_sync_file(self, content: str, path: str):
        if not self.compress:
            with open(path, "w") as fp:
                fp.write(content)
        else:  # lzma
            compressed_data = lzma.compress(content.encode())
            with open(path, "wb") as compress_fp:
                compress_fp.write(compressed_data)

    def save_json(self, data: dict, *names: Any):
        file_name = self._get_json_file_name(names)
        folder_path = self.folder_path
        os.makedirs(folder_path, exist_ok=True)
        path = folder_path + file_name
        content = json.dumps(data, ensure_ascii=False)
        self._save_sync_file(content, path)
        self.logger.debug(f"success saved: {path}")
        return path

    def save_html(self, content: str, *names: Any):
        file_name = self._get_html_file_name(names)
        folder_path = self.folder_path
        os.makedirs(folder_path, exist_ok=True)
        path = folder_path + file_name
        self._save_sync_file(content, path)
        self.logger.debug(f"success saved: {path}")
        return path
